/*
 * NT threads support
 *
 * Copyright 1996, 2003 Alexandre Julliard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include "wine/port.h"

#include <sys/types.h>
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#include "ntstatus.h"
#include "thread.h"
#include "winternl.h"
#include "wine/library.h"
#include "wine/server.h"
#include "wine/pthread.h"
#include "wine/debug.h"
#include "ntdll_misc.h"

WINE_DEFAULT_DEBUG_CHANNEL(thread);

/* info passed to a starting thread */
struct startup_info
{
    struct wine_pthread_thread_info pthread_info;
    PRTL_THREAD_START_ROUTINE       entry_point;
    void                           *entry_arg;
};

static PEB peb;
static PEB_LDR_DATA ldr;
static RTL_USER_PROCESS_PARAMETERS params;  /* default parameters if no parent */
static RTL_BITMAP tls_bitmap;
static LIST_ENTRY tls_links;


/***********************************************************************
 *           alloc_teb
 */
static TEB *alloc_teb( ULONG *size )
{
    TEB *teb;

    *size = SIGNAL_STACK_SIZE + sizeof(TEB);
    teb = wine_anon_mmap( NULL, *size, PROT_READ | PROT_WRITE | PROT_EXEC, 0 );
    if (teb == (TEB *)-1) return NULL;
    if (!(teb->teb_sel = wine_ldt_alloc_fs()))
    {
        munmap( teb, *size );
        return NULL;
    }
    teb->Tib.ExceptionList = (void *)~0UL;
    teb->Tib.StackBase     = (void *)~0UL;
    teb->Tib.Self          = &teb->Tib;
    teb->Peb               = &peb;
    teb->StaticUnicodeString.Buffer        = teb->StaticUnicodeBuffer;
    teb->StaticUnicodeString.MaximumLength = sizeof(teb->StaticUnicodeBuffer);
    teb->ThreadLocalStoragePointer = teb->TlsSlots;
    return teb;
}


/***********************************************************************
 *           free_teb
 */
static inline void free_teb( TEB *teb )
{
    ULONG size = 0;
    void *addr = teb;

    NtFreeVirtualMemory( GetCurrentProcess(), &addr, &size, MEM_RELEASE );
    wine_ldt_free_fs( teb->teb_sel );
    munmap( teb, SIGNAL_STACK_SIZE + sizeof(TEB) );
}


/***********************************************************************
 *           thread_init
 *
 * Setup the initial thread.
 *
 * NOTES: The first allocated TEB on NT is at 0x7ffde000.
 */
void thread_init(void)
{
    TEB *teb;
    void *addr;
    ULONG size;
    struct wine_pthread_thread_info thread_info;
    static struct debug_info debug_info;  /* debug info for initial thread */

    peb.NumberOfProcessors = 1;
    peb.ProcessParameters  = &params;
    peb.TlsBitmap          = &tls_bitmap;
    peb.LdrData            = &ldr;
    RtlInitializeBitMap( &tls_bitmap, (BYTE *)peb.TlsBitmapBits, sizeof(peb.TlsBitmapBits) * 8 );
    InitializeListHead( &ldr.InLoadOrderModuleList );
    InitializeListHead( &ldr.InMemoryOrderModuleList );
    InitializeListHead( &ldr.InInitializationOrderModuleList );
    InitializeListHead( &tls_links );

    teb = alloc_teb( &size );
    teb->request_fd    = -1;
    teb->reply_fd      = -1;
    teb->wait_fd[0]    = -1;
    teb->wait_fd[1]    = -1;
    teb->debug_info    = &debug_info;
    InsertHeadList( &tls_links, &teb->TlsLinks );

    thread_info.stack_base = NULL;
    thread_info.stack_size = 0;
    thread_info.teb_base   = teb;
    thread_info.teb_size   = size;
    thread_info.teb_sel    = teb->teb_sel;
    wine_pthread_init_current_teb( &thread_info );
    wine_pthread_init_thread( &thread_info );

    debug_info.str_pos = debug_info.strings;
    debug_info.out_pos = debug_info.output;
    debug_init();
    virtual_init();

    /* setup the server connection */
    server_init_process();
    server_init_thread( thread_info.pid, thread_info.tid, NULL );

    /* create a memory view for the TEB */
    NtAllocateVirtualMemory( GetCurrentProcess(), &addr, teb, &size,
                             MEM_SYSTEM, PAGE_EXECUTE_READWRITE );

    /* create the process heap */
    if (!(peb.ProcessHeap = RtlCreateHeap( HEAP_GROWABLE, NULL, 0, 0, NULL, NULL )))
    {
        MESSAGE( "wine: failed to create the process heap\n" );
        exit(1);
    }
}


/***********************************************************************
 *           start_thread
 *
 * Startup routine for a newly created thread.
 */
static void start_thread( struct wine_pthread_thread_info *info )
{
    TEB *teb = info->teb_base;
    struct startup_info *startup_info = (struct startup_info *)info;
    PRTL_THREAD_START_ROUTINE func = startup_info->entry_point;
    void *arg = startup_info->entry_arg;
    struct debug_info debug_info;
    ULONG size;

    debug_info.str_pos = debug_info.strings;
    debug_info.out_pos = debug_info.output;
    teb->debug_info = &debug_info;

    wine_pthread_init_current_teb( info );
    SIGNAL_Init();
    server_init_thread( info->pid, info->tid, func );
    wine_pthread_init_thread( info );

    /* allocate a memory view for the stack */
    size = info->stack_size;
    NtAllocateVirtualMemory( GetCurrentProcess(), &teb->DeallocationStack, info->stack_base,
                             &size, MEM_SYSTEM, PAGE_EXECUTE_READWRITE );
    /* limit is lower than base since the stack grows down */
    teb->Tib.StackBase  = (char *)info->stack_base + info->stack_size;
    teb->Tib.StackLimit = info->stack_base;

    /* setup the guard page */
    size = 1;
    NtProtectVirtualMemory( GetCurrentProcess(), &teb->DeallocationStack, &size,
                            PAGE_EXECUTE_READWRITE | PAGE_GUARD, NULL );
    RtlFreeHeap( GetProcessHeap(), 0, info );

    RtlAcquirePebLock();
    InsertHeadList( &tls_links, &teb->TlsLinks );
    RtlReleasePebLock();

    func( arg );
}


/***********************************************************************
 *              RtlCreateUserThread   (NTDLL.@)
 */
NTSTATUS WINAPI RtlCreateUserThread( HANDLE process, const SECURITY_DESCRIPTOR *descr,
                                     BOOLEAN suspended, PVOID stack_addr,
                                     SIZE_T stack_reserve, SIZE_T stack_commit,
                                     PRTL_THREAD_START_ROUTINE start, void *param,
                                     HANDLE *handle_ptr, CLIENT_ID *id )
{
    struct startup_info *info = NULL;
    HANDLE handle = 0;
    TEB *teb = NULL;
    DWORD tid = 0;
    ULONG size;
    int request_pipe[2];
    NTSTATUS status;

    if (pipe( request_pipe ) == -1) return STATUS_TOO_MANY_OPENED_FILES;
    fcntl( request_pipe[1], F_SETFD, 1 ); /* set close on exec flag */
    wine_server_send_fd( request_pipe[0] );

    SERVER_START_REQ( new_thread )
    {
        req->suspend    = suspended;
        req->inherit    = 0;  /* FIXME */
        req->request_fd = request_pipe[0];
        if (!(status = wine_server_call( req )))
        {
            handle = reply->handle;
            tid = reply->tid;
        }
        close( request_pipe[0] );
    }
    SERVER_END_REQ;

    if (status) goto error;

    if (!(info = RtlAllocateHeap( GetProcessHeap(), 0, sizeof(*info) )))
    {
        status = STATUS_NO_MEMORY;
        goto error;
    }

    if (!(teb = alloc_teb( &size )))
    {
        status = STATUS_NO_MEMORY;
        goto error;
    }
    teb->ClientId.UniqueProcess = (HANDLE)GetCurrentProcessId();
    teb->ClientId.UniqueThread  = (HANDLE)tid;

    teb->exit_code   = STILL_ACTIVE;
    teb->request_fd  = request_pipe[1];
    teb->reply_fd    = -1;
    teb->wait_fd[0]  = -1;
    teb->wait_fd[1]  = -1;
    teb->htask16     = NtCurrentTeb()->htask16;

    NtAllocateVirtualMemory( GetCurrentProcess(), &info->pthread_info.teb_base, teb, &size,
                             MEM_SYSTEM, PAGE_EXECUTE_READWRITE );
    info->pthread_info.teb_size = size;
    info->pthread_info.teb_sel  = teb->teb_sel;

    if (!stack_reserve || !stack_commit)
    {
        IMAGE_NT_HEADERS *nt = RtlImageNtHeader( NtCurrentTeb()->Peb->ImageBaseAddress );
        if (!stack_reserve) stack_reserve = nt->OptionalHeader.SizeOfStackReserve;
        if (!stack_commit) stack_commit = nt->OptionalHeader.SizeOfStackCommit;
    }
    if (stack_reserve < stack_commit) stack_reserve = stack_commit;
    stack_reserve = (stack_reserve + 0xffff) & ~0xffff;  /* round to 64K boundary */
    if (stack_reserve < 1024 * 1024) stack_reserve = 1024 * 1024;  /* Xlib needs a large stack */

    info->pthread_info.stack_base = NULL;
    info->pthread_info.stack_size = stack_reserve;
    info->pthread_info.entry      = start_thread;
    info->entry_point             = start;
    info->entry_arg               = param;

    if (wine_pthread_create_thread( &info->pthread_info ) == -1)
    {
        status = STATUS_NO_MEMORY;
        goto error;
    }

    if (id) id->UniqueThread = (HANDLE)tid;
    if (handle_ptr) *handle_ptr = handle;
    else NtClose( handle );

    return STATUS_SUCCESS;

error:
    if (teb) free_teb( teb );
    if (info) RtlFreeHeap( GetProcessHeap(), 0, info );
    if (handle) NtClose( handle );
    close( request_pipe[1] );
    return status;
}


/***********************************************************************
 *              NtOpenThread   (NTDLL.@)
 *              ZwOpenThread   (NTDLL.@)
 */
NTSTATUS WINAPI NtOpenThread( HANDLE *handle, ACCESS_MASK access,
                              const OBJECT_ATTRIBUTES *attr, const CLIENT_ID *id )
{
    NTSTATUS ret;

    SERVER_START_REQ( open_thread )
    {
        req->tid     = (thread_id_t)id->UniqueThread;
        req->access  = access;
        req->inherit = attr && (attr->Attributes & OBJ_INHERIT);
        ret = wine_server_call( req );
        *handle = reply->handle;
    }
    SERVER_END_REQ;
    return ret;
}


/******************************************************************************
 *              NtSuspendThread   (NTDLL.@)
 *              ZwSuspendThread   (NTDLL.@)
 */
NTSTATUS WINAPI NtSuspendThread( HANDLE handle, PULONG count )
{
    NTSTATUS ret;

    SERVER_START_REQ( suspend_thread )
    {
        req->handle = handle;
        if (!(ret = wine_server_call( req ))) *count = reply->count;
    }
    SERVER_END_REQ;
    return ret;
}


/******************************************************************************
 *              NtResumeThread   (NTDLL.@)
 *              ZwResumeThread   (NTDLL.@)
 */
NTSTATUS WINAPI NtResumeThread( HANDLE handle, PULONG count )
{
    NTSTATUS ret;

    SERVER_START_REQ( resume_thread )
    {
        req->handle = handle;
        if (!(ret = wine_server_call( req ))) *count = reply->count;
    }
    SERVER_END_REQ;
    return ret;
}


/******************************************************************************
 *              NtTerminateThread  (NTDLL.@)
 *              ZwTerminateThread  (NTDLL.@)
 */
NTSTATUS WINAPI NtTerminateThread( HANDLE handle, LONG exit_code )
{
    NTSTATUS ret;
    BOOL self, last;

    SERVER_START_REQ( terminate_thread )
    {
        req->handle    = handle;
        req->exit_code = exit_code;
        ret = wine_server_call( req );
        self = !ret && reply->self;
        last = reply->last;
    }
    SERVER_END_REQ;

    if (self)
    {
        if (last) exit( exit_code );
        else server_abort_thread( exit_code );
    }
    return ret;
}


/******************************************************************************
 *              NtQueueApcThread  (NTDLL.@)
 */
NTSTATUS WINAPI NtQueueApcThread( HANDLE handle, PNTAPCFUNC func, ULONG_PTR arg1,
                                  ULONG_PTR arg2, ULONG_PTR arg3 )
{
    NTSTATUS ret;
    SERVER_START_REQ( queue_apc )
    {
        req->handle = handle;
        req->user   = 1;
        req->func   = func;
        req->arg1   = (void *)arg1;
        req->arg2   = (void *)arg2;
        req->arg3   = (void *)arg3;
        ret = wine_server_call( req );
    }
    SERVER_END_REQ;
    return ret;
}


/***********************************************************************
 *              NtSetContextThread  (NTDLL.@)
 *              ZwSetContextThread  (NTDLL.@)
 */
NTSTATUS WINAPI NtSetContextThread( HANDLE handle, const CONTEXT *context )
{
    NTSTATUS ret;

    SERVER_START_REQ( set_thread_context )
    {
        req->handle  = handle;
        req->flags   = context->ContextFlags;
        req->suspend = 0;
        wine_server_add_data( req, context, sizeof(*context) );
        ret = wine_server_call( req );
    }
    SERVER_END_REQ;
    return ret;
}


/***********************************************************************
 *              NtGetContextThread  (NTDLL.@)
 *              ZwGetContextThread  (NTDLL.@)
 */
NTSTATUS WINAPI NtGetContextThread( HANDLE handle, CONTEXT *context )
{
    NTSTATUS ret;
    CONTEXT ctx;

    SERVER_START_REQ( get_thread_context )
    {
        req->handle  = handle;
        req->flags   = context->ContextFlags;
        req->suspend = 0;
        wine_server_set_reply( req, &ctx, sizeof(ctx) );
        ret = wine_server_call( req );
    }
    SERVER_END_REQ;

    if (ret) return ret;

#ifdef __i386__
    if (context->ContextFlags & CONTEXT_INTEGER)
    {
        context->Eax = ctx.Eax;
        context->Ebx = ctx.Ebx;
        context->Ecx = ctx.Ecx;
        context->Edx = ctx.Edx;
        context->Esi = ctx.Esi;
        context->Edi = ctx.Edi;
    }
    if (context->ContextFlags & CONTEXT_CONTROL)
    {
        context->Ebp    = ctx.Ebp;
        context->Esp    = ctx.Esp;
        context->Eip    = ctx.Eip;
        context->SegCs  = ctx.SegCs;
        context->SegSs  = ctx.SegSs;
        context->EFlags = ctx.EFlags;
    }
    if (context->ContextFlags & CONTEXT_SEGMENTS)
    {
        context->SegDs = ctx.SegDs;
        context->SegEs = ctx.SegEs;
        context->SegFs = ctx.SegFs;
        context->SegGs = ctx.SegGs;
    }
    if (context->ContextFlags & CONTEXT_DEBUG_REGISTERS)
    {
        context->Dr0 = ctx.Dr0;
        context->Dr1 = ctx.Dr1;
        context->Dr2 = ctx.Dr2;
        context->Dr3 = ctx.Dr3;
        context->Dr6 = ctx.Dr6;
        context->Dr7 = ctx.Dr7;
    }
    if (context->ContextFlags & CONTEXT_FLOATING_POINT)
    {
        context->FloatSave = ctx.FloatSave;
    }
#elif defined(__sparc__)
    if (context->ContextFlags & CONTEXT_CONTROL)
    {
        context->psr = ctx.psr;
        context->pc  = ctx.pc;
        context->npc = ctx.npc;
        context->y   = ctx.y;
        context->wim = ctx.wim;
        context->tbr = ctx.tbr;
    }
    if (context->ContextFlags & CONTEXT_INTEGER)
    {
        context->g0 = ctx.g0;
        context->g1 = ctx.g1;
        context->g2 = ctx.g2;
        context->g3 = ctx.g3;
        context->g4 = ctx.g4;
        context->g5 = ctx.g5;
        context->g6 = ctx.g6;
        context->g7 = ctx.g7;
        context->o0 = ctx.o0;
        context->o1 = ctx.o1;
        context->o2 = ctx.o2;
        context->o3 = ctx.o3;
        context->o4 = ctx.o4;
        context->o5 = ctx.o5;
        context->o6 = ctx.o6;
        context->o7 = ctx.o7;
        context->l0 = ctx.l0;
        context->l1 = ctx.l1;
        context->l2 = ctx.l2;
        context->l3 = ctx.l3;
        context->l4 = ctx.l4;
        context->l5 = ctx.l5;
        context->l6 = ctx.l6;
        context->l7 = ctx.l7;
        context->i0 = ctx.i0;
        context->i1 = ctx.i1;
        context->i2 = ctx.i2;
        context->i3 = ctx.i3;
        context->i4 = ctx.i4;
        context->i5 = ctx.i5;
        context->i6 = ctx.i6;
        context->i7 = ctx.i7;
    }
    if (context->ContextFlags & CONTEXT_FLOATING_POINT)
    {
        /* FIXME */
    }
#elif defined(__powerpc__)
    if (context->ContextFlags & CONTEXT_CONTROL)
    {
        context->Msr = ctx.Msr;
        context->Ctr = ctx.Ctr;
        context->Iar = ctx.Iar;
    }
    if (context->ContextFlags & CONTEXT_INTEGER)
    {
        context->Gpr0  = ctx.Gpr0;
        context->Gpr1  = ctx.Gpr1;
        context->Gpr2  = ctx.Gpr2;
        context->Gpr3  = ctx.Gpr3;
        context->Gpr4  = ctx.Gpr4;
        context->Gpr5  = ctx.Gpr5;
        context->Gpr6  = ctx.Gpr6;
        context->Gpr7  = ctx.Gpr7;
        context->Gpr8  = ctx.Gpr8;
        context->Gpr9  = ctx.Gpr9;
        context->Gpr10 = ctx.Gpr10;
        context->Gpr11 = ctx.Gpr11;
        context->Gpr12 = ctx.Gpr12;
        context->Gpr13 = ctx.Gpr13;
        context->Gpr14 = ctx.Gpr14;
        context->Gpr15 = ctx.Gpr15;
        context->Gpr16 = ctx.Gpr16;
        context->Gpr17 = ctx.Gpr17;
        context->Gpr18 = ctx.Gpr18;
        context->Gpr19 = ctx.Gpr19;
        context->Gpr20 = ctx.Gpr20;
        context->Gpr21 = ctx.Gpr21;
        context->Gpr22 = ctx.Gpr22;
        context->Gpr23 = ctx.Gpr23;
        context->Gpr24 = ctx.Gpr24;
        context->Gpr25 = ctx.Gpr25;
        context->Gpr26 = ctx.Gpr26;
        context->Gpr27 = ctx.Gpr27;
        context->Gpr28 = ctx.Gpr28;
        context->Gpr29 = ctx.Gpr29;
        context->Gpr30 = ctx.Gpr30;
        context->Gpr31 = ctx.Gpr31;
        context->Xer   = ctx.Xer;
        context->Cr    = ctx.Cr;
    }
    if (context->ContextFlags & CONTEXT_FLOATING_POINT)
    {
        context->Fpr0  = ctx.Fpr0;
        context->Fpr1  = ctx.Fpr1;
        context->Fpr2  = ctx.Fpr2;
        context->Fpr3  = ctx.Fpr3;
        context->Fpr4  = ctx.Fpr4;
        context->Fpr5  = ctx.Fpr5;
        context->Fpr6  = ctx.Fpr6;
        context->Fpr7  = ctx.Fpr7;
        context->Fpr8  = ctx.Fpr8;
        context->Fpr9  = ctx.Fpr9;
        context->Fpr10 = ctx.Fpr10;
        context->Fpr11 = ctx.Fpr11;
        context->Fpr12 = ctx.Fpr12;
        context->Fpr13 = ctx.Fpr13;
        context->Fpr14 = ctx.Fpr14;
        context->Fpr15 = ctx.Fpr15;
        context->Fpr16 = ctx.Fpr16;
        context->Fpr17 = ctx.Fpr17;
        context->Fpr18 = ctx.Fpr18;
        context->Fpr19 = ctx.Fpr19;
        context->Fpr20 = ctx.Fpr20;
        context->Fpr21 = ctx.Fpr21;
        context->Fpr22 = ctx.Fpr22;
        context->Fpr23 = ctx.Fpr23;
        context->Fpr24 = ctx.Fpr24;
        context->Fpr25 = ctx.Fpr25;
        context->Fpr26 = ctx.Fpr26;
        context->Fpr27 = ctx.Fpr27;
        context->Fpr28 = ctx.Fpr28;
        context->Fpr29 = ctx.Fpr29;
        context->Fpr30 = ctx.Fpr30;
        context->Fpr31 = ctx.Fpr31;
        context->Fpscr = ctx.Fpscr;
    }
#else
#error You must implement context copying for your CPU
#endif
    return STATUS_SUCCESS;
}


/******************************************************************************
 *              NtQueryInformationThread  (NTDLL.@)
 *              ZwQueryInformationThread  (NTDLL.@)
 */
NTSTATUS WINAPI NtQueryInformationThread( HANDLE handle, THREADINFOCLASS class,
                                          void *data, ULONG length, ULONG *ret_len )
{
    NTSTATUS status;

    switch(class)
    {
    case ThreadBasicInformation:
        {
            THREAD_BASIC_INFORMATION info;

            SERVER_START_REQ( get_thread_info )
            {
                req->handle = handle;
                req->tid_in = 0;
                if (!(status = wine_server_call( req )))
                {
                    info.ExitStatus             = reply->exit_code;
                    info.TebBaseAddress         = reply->teb;
                    info.ClientId.UniqueProcess = (HANDLE)reply->pid;
                    info.ClientId.UniqueThread  = (HANDLE)reply->tid;
                    info.AffinityMask           = reply->affinity;
                    info.Priority               = reply->priority;
                    info.BasePriority           = reply->priority;  /* FIXME */
                }
            }
            SERVER_END_REQ;
            if (status == STATUS_SUCCESS)
            {
                if (data) memcpy( data, &info, min( length, sizeof(info) ));
                if (ret_len) *ret_len = min( length, sizeof(info) );
            }
        }
        return status;
    case ThreadTimes:
    case ThreadPriority:
    case ThreadBasePriority:
    case ThreadAffinityMask:
    case ThreadImpersonationToken:
    case ThreadDescriptorTableEntry:
    case ThreadEnableAlignmentFaultFixup:
    case ThreadEventPair_Reusable:
    case ThreadQuerySetWin32StartAddress:
    case ThreadZeroTlsCell:
    case ThreadPerformanceCount:
    case ThreadAmILastThread:
    case ThreadIdealProcessor:
    case ThreadPriorityBoost:
    case ThreadSetTlsArrayAddress:
    case ThreadIsIoPending:
    default:
        FIXME( "info class %d not supported yet\n", class );
        return STATUS_NOT_IMPLEMENTED;
    }
}


/******************************************************************************
 *              NtSetInformationThread  (NTDLL.@)
 *              ZwSetInformationThread  (NTDLL.@)
 */
NTSTATUS WINAPI NtSetInformationThread( HANDLE handle, THREADINFOCLASS class,
                                        LPCVOID data, ULONG length )
{
    switch(class)
    {
    case ThreadZeroTlsCell:
        if (handle == GetCurrentThread())
        {
            LIST_ENTRY *entry;
            DWORD index;

            if (length != sizeof(DWORD)) return STATUS_INVALID_PARAMETER;
            index = *(const DWORD *)data;
            if (index >= 64) return STATUS_INVALID_PARAMETER;
            RtlAcquirePebLock();
            for (entry = tls_links.Flink; entry != &tls_links; entry = entry->Flink)
            {
                TEB *teb = CONTAINING_RECORD(entry, TEB, TlsLinks);
                teb->TlsSlots[index] = 0;
            }
            RtlReleasePebLock();
            return STATUS_SUCCESS;
        }
        FIXME( "ZeroTlsCell not supported on other threads\n" );
        return STATUS_NOT_IMPLEMENTED;

    case ThreadBasicInformation:
    case ThreadTimes:
    case ThreadPriority:
    case ThreadBasePriority:
    case ThreadAffinityMask:
    case ThreadImpersonationToken:
    case ThreadDescriptorTableEntry:
    case ThreadEnableAlignmentFaultFixup:
    case ThreadEventPair_Reusable:
    case ThreadQuerySetWin32StartAddress:
    case ThreadPerformanceCount:
    case ThreadAmILastThread:
    case ThreadIdealProcessor:
    case ThreadPriorityBoost:
    case ThreadSetTlsArrayAddress:
    case ThreadIsIoPending:
    default:
        FIXME( "info class %d not supported yet\n", class );
        return STATUS_NOT_IMPLEMENTED;
    }
}


/**********************************************************************
 *           NtCurrentTeb   (NTDLL.@)
 */
#if defined(__i386__) && defined(__GNUC__)

__ASM_GLOBAL_FUNC( NtCurrentTeb, ".byte 0x64\n\tmovl 0x18,%eax\n\tret" );

#elif defined(__i386__) && defined(_MSC_VER)

/* Nothing needs to be done. MS C "magically" exports the inline version from winnt.h */

#else

/**********************************************************************/

TEB * WINAPI NtCurrentTeb(void)
{
    return wine_pthread_get_current_teb();
}

#endif  /* __i386__ */
